// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::env;

use png::ColorType;

use super::image::Image;
use super::inkscape;

use crate::utils;

fn gen_preview(input: &Image, output: &mut Image, player_colour: [u8; 3], x_off: u32, y_off: u32) {
	for y in 0..input.height {
		for x in 0..input.width {
			let src_i = (x + y * input.width) as usize;
			let base = input.data[src_i * 2];
			let colour = input.data[src_i * 2 + 1];
			let t = colour as f64 / 255.0;
			for dst_i in [((x + x_off) + (y + y_off) * output.width) as usize, ((x + x_off) + ((input.width - 1 - y) + y_off) * output.width) as usize] {
				for (i, &colour) in player_colour.iter().enumerate() { output.data[dst_i * 4 + i] = (base as i32 + ((colour as i32 - base as i32) as f64 * t) as i32) as u8; }
				output.data[dst_i * 4 + 3] = 255;
			}
		}
	}
}

pub fn generate() {
	const INPUT: &str = "data/player.svg";
	const OUTPUT: &str = "../data/textures/player.png";

	const BASE_PATH: &str = "/tmp/base.png";
	const COLOUR_PATH: &str = "/tmp/colour.png";

	const WIDTH: u32 = 128;
	const HEIGHT: u32 = WIDTH / 2;

	if utils::newer(INPUT, OUTPUT) {
		inkscape::gen_layer_grey(INPUT, BASE_PATH, "base", WIDTH, HEIGHT);
		inkscape::gen_layer_grey(INPUT, COLOUR_PATH, "colour", WIDTH, HEIGHT);

		let (width, height, base, colour) = {
			let base = Image::load(BASE_PATH, ColorType::Grayscale);
			let colour = Image::load(COLOUR_PATH, ColorType::Grayscale);

			assert!(base.width == colour.width && base.height == colour.height, "Base and colour layers aren't the same dimensions");
			(base.width, base.height, base.data, colour.data)
		};

		let mut output = Image::new(width, height, 2);
		for i in 0..(width * height) as usize {
			(output.data[i * 2], output.data[i * 2 + 1]) = (base[i], colour[i]);
		}
		output.save(OUTPUT, ColorType::GrayscaleAlpha);

		if env::args().nth(1).as_deref() == Some("preview") {
			let mut preview = Image::new(width * 3, (height * 2) * 3, 4);

			let colours = [
				[255, 63, 31],
				[31, 63, 255],
				[63, 255, 31],
				[255, 255, 31],
				[31, 255, 255],
				[255, 31, 255],
				[255, 255, 255],
				[127, 127, 127],
			];

			for (i, &colour) in colours.iter().enumerate() {
				let i = i as u32;
				let (x_off, y_off) = ((i % 3) * output.width, (i / 3) * output.width);
				gen_preview(&output, &mut preview, colour, x_off, y_off);
			}
			preview.save("/tmp/preview.png", ColorType::Rgba);
		}
	}
}
