// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use glam::{UVec2, Vec2, Mat4};
use strum::EnumCount;
use strum_macros::EnumCount as EnumCountMacro;

use crate::game::config::action::{Action, ActionEvent, ActionState};

#[derive(EnumCountMacro)]
enum ZoomAction { In, Out }

#[derive(Default)]
pub struct Camera {
	pos: Vec2,
	zoom: f32,
	zoom_rendered: f32,
	size: Vec2,
	winsize: Vec2,
	matrix: Mat4,
	zoom_actions: [bool; ZoomAction::COUNT],
}

const ZOOM_MIN: f32 = 4.0;
const ZOOM_MAX: f32 = 100.0;
const ZOOM_MAX_LEVEL: f32 = 20.0;
const INIT_ZOOM: f32 = ZOOM_MAX_LEVEL / 2.0;

impl Camera {
	pub fn new(winsize: UVec2) -> Camera {
		let zoom = INIT_ZOOM;
		Camera {
			zoom,
			zoom_rendered: zoom,
			winsize: winsize.as_vec2(),
			.. Camera::default()
		}
	}

	pub fn reset(&mut self, player_pos: Vec2, player_vel: Vec2, winsize: UVec2) {
		self.pos = player_pos;
		self.zoom = INIT_ZOOM;
		self.zoom_rendered = INIT_ZOOM;
		self.winsize = winsize.as_vec2();
		self.update(player_pos, player_vel, 0.0);
	}

	pub fn get_pos(&self) -> Vec2 { self.pos }
	pub fn get_size(&self) -> Vec2 { self.size }
	pub fn get_matrix(&self) -> &Mat4 { &self.matrix }

	pub fn resize_event(&mut self, winsize: UVec2) {
		self.winsize = winsize.as_vec2();
	}

	pub fn mouse_scroll_event(&mut self, dy: f32) {
		self.change_zoom(-dy);
	}

	pub fn get_window_size(&self) -> UVec2 {
		self.winsize.as_uvec2()
	}

	fn change_zoom(&mut self, diff: f32) {
		self.zoom = (self.zoom + diff).clamp(0.0, ZOOM_MAX_LEVEL);
	}

	pub fn action(&mut self, action: ActionEvent) {
		match action {
			ActionEvent::Action(Action::ZoomReset, ActionState::Pressed) => self.zoom = INIT_ZOOM,
			ActionEvent::Action(Action::ZoomIn, state) => self.zoom_actions[ZoomAction::In as usize] = state.pressed(),
			ActionEvent::Action(Action::ZoomOut, state) => self.zoom_actions[ZoomAction::Out as usize] = state.pressed(),
			ActionEvent::Action(..) => (),
			ActionEvent::AllReleased => self.zoom_actions = Default::default(),
		}
	}

	pub fn move_by(&mut self, dpos: Vec2) {
		self.pos += dpos;
	}

	pub fn pixel_to_world_coords(&self, pos: Vec2) -> Vec2 {
		self.pos + self.size * Vec2::new(
			pos.x / self.winsize.x - 0.5,
			0.5 - pos.y / self.winsize.y,
		)
	}

	pub fn world_to_pixel_coords(&self, pos: Vec2) -> Vec2 {
		let half_ndc = (pos - self.pos) / self.size; // Both x and y between [-0.5, 0.5]
		Vec2::new(
			(half_ndc.x + 0.5) * self.winsize.x,
			(0.5 - half_ndc.y) * self.winsize.y,
		)
	}

	pub fn update(&mut self, player_pos: Vec2, player_vel: Vec2, dt: f32) -> Vec2 {
		let mut dir = 0;
		if self.zoom_actions[ZoomAction::Out as usize] { dir += 1; }
		if self.zoom_actions[ZoomAction::In as usize] { dir -= 1; }
		self.change_zoom(dir as f32 * dt * 12.0);

		let diff_centre = (player_pos - self.pos) / Camera::calc_zoom(self.zoom_rendered);

		// Exponentially decays towards the real zoom, making it smooth
		let diff = self.zoom - self.zoom_rendered;
		self.zoom_rendered += diff * dt * 12.0;

		// Prevents exponential decay from resulting in things flipping
		if (self.zoom - self.zoom_rendered) * diff < 0.0 {
			self.zoom_rendered = self.zoom;
		}

		let zoom = Camera::calc_zoom(self.zoom_rendered);
		self.pos = player_pos - diff_centre * zoom;

		let rate = 128.0 / zoom;

		/*
		 * Okay, this is complicated.
		 *
		 * One problem that I encountered when programming this game is that the
		 * camera's position jitters a bit when the frame rate varies due to lag.
		 *
		 * Previously I was having the camera's position exponentially decay to
		 * the player's position by using a linear approximation of the
		 * differential equation `dy = -ky dt`. This approximation is just
		 * calculating `y -= k * dt`.
		 *
		 * This is cheap to calculate but very inaccurate. The inaccuracy isn't a
		 * problem by itself as this is just an animation, but when the delta time
		 * varies and the player has been accelerating in some direction for a
		 * while (so the camera's position relative to the player is constant),
		 * small changes in the delta time result in small changes in where this
		 * maximum difference should be, which "jitters" the camera a bit.
		 *
		 * The solution to this is to not use the approximation and instead use
		 * the accurate calculation. This is what the following line of code does.
		 * This is the solution to the differential equation
		 * 	dc = -k(c - p) dt
		 * which is:
		 * 	c = p_0 + v(t - 1 / k) + (c_0 - p_0 + v / k)e^(-kt)
		 * with t = dt.
		 *
		 * This differential equation is basically simulating the camera
		 * exponentially decaying to the *moving* player.
		 */
		let prev_pos = self.pos;
		self.pos = player_pos + player_vel * (dt - 1.0 / rate) + (self.pos - player_pos + player_vel / rate) * (-rate * dt).exp();

		self.size = if self.winsize.x > self.winsize.y {
			Vec2::new(zoom * self.winsize.x / self.winsize.y, zoom)
		} else {
			Vec2::new(zoom, zoom * self.winsize.y / self.winsize.x)
		};

		self.matrix = Mat4::orthographic_rh_gl(
			self.pos.x - self.size.x / 2.0, self.pos.x + self.size.x / 2.0,
			self.pos.y - self.size.y / 2.0, self.pos.y + self.size.y / 2.0,
			-1.0, 1.0,
		);

		self.pos - prev_pos
	}

	fn calc_zoom(rendered: f32) -> f32 {
		ZOOM_MIN * (ZOOM_MAX / ZOOM_MIN).powf(rendered / ZOOM_MAX_LEVEL)
	}
}
