// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod action;
mod action_manager;

use std::{fs::File, path::Path};
use std::io::{Read, Write, BufReader, BufWriter, ErrorKind};

use serde::{Serialize, Deserialize};
use ron::{ser::{self, PrettyConfig}, de};
use fs4::fs_std::FileExt;

use action::ActionManager;

use super::Filesystem;

#[derive(Serialize, Deserialize)]
pub struct Config {
	#[serde(rename = "input_bindings")] pub action_manager: ActionManager,
	pub graphics: Graphics,
}

#[derive(Serialize, Deserialize)]
pub struct Graphics {
	pub star_density: f32,
	pub particles: Particles,
}

#[derive(Serialize, Deserialize)]
pub struct Particles {
	pub thrust: bool,
	pub powerup: bool,
	pub death: bool,
	pub hit: bool,
}

const PATH: &str = "config.ron";

impl Config {
	pub fn new(fs: &Filesystem) -> Config {
		Config::load(fs).unwrap_or_else(|err| {
			let ret = Config::default();
			if let Some(err) = err {
				log::warn!("failed loading config: {err}");
				/*
				 * Don't want to save the config file on error, as it might be a
				 * syntax error from manual editing and I don't want to overwrite
				 * the user's changes.
				 */
			} else {
				log::info!("Config file doesn't exist, creating one");
				if let Err(err) = ret.save(fs) {
					log::warn!("cannot save config: {err}");
				}
			}

			ret
		})
	}

	/**
	 * Tries to load the config from the file.
	 *
	 * Returns `Ok` on success, `Err(None)` when the file isn't found (so falling
	 * back to default) or `Err(Some(String))` on another error worth logging.
	 */
	fn load(fs: &Filesystem) -> Result<Config, Option<String>> {
		let path = fs.get(Path::new(PATH));
		let file = match File::open(&path) {
			Ok(file) => file,
			Err(err) if err.kind() == ErrorKind::NotFound => return Err(None),
			Err(err) => return Err(Some(format!("cannot open file \"{}\": {err}", path.display()))),
		};
		file.lock_shared().map_err(|err| Some(format!("cannot lock (shared) file: {err}")))?;
		let ret = Config::deserialise(BufReader::new(&file));
		file.unlock().map_err(|err| Some(format!("cannot unlock file: {err}")))?;
		ret
	}

	fn deserialise(reader: impl Read) -> Result<Config, Option<String>> {
		de::from_reader(reader).map_err(|err| Some(format!("failed deserialising: {err}")))
	}

	pub fn save(&self, fs: &Filesystem) -> Result<(), String> {
		/*
		 * Cannot call `File::create` because that truncates the file, which
		 * breaks the locking. Instead I am truncating the file but only after
		 * locking it.
		 */
		let path = fs.get(Path::new(PATH));
		let file = File::options().write(true).create(true).truncate(false).open(&path).map_err(|err| format!("cannot create file \"{}\": {err}", path.display()))?;
		file.lock_exclusive().map_err(|err| format!("cannot lock (exclusive) file: {err}"))?;
		file.set_len(0).map_err(|err| format!("cannot truncate file: {err}"))?;
		let ret = self.serialise(BufWriter::new(&file));
		file.unlock().map_err(|err| format!("cannot unlock file: {err}"))?;
		ret
	}

	fn serialise(&self, mut writer: impl Write) -> Result<(), String> {
		writeln!(writer, "// Warning: All comments made will be automatically overwritten when changing these in the GUI").map_err(|err| format!("cannot write first comment: {err}"))?;
		ser::to_writer_pretty(&mut writer, self, PrettyConfig::new().indentor(String::from("\t"))).map_err(|err| format!("cannot serialise: {err}"))?;
		writeln!(writer).map_err(|err| format!("cannot write final newline: {err}"))?; // Need to add a final newline
		writer.flush().map_err(|err| format!("cannot flush file: {err}"))
	}
}

impl Default for Config {
	fn default() -> Config {
		Config {
			action_manager: ActionManager::default(),
			graphics: Graphics {
				star_density: DEFAULT_STAR_DENSITY,
				particles: Particles { thrust: true, powerup: true, death: true, hit: true },
			},
		}
	}
}

pub const DEFAULT_STAR_DENSITY: f32 = 800.0;
