// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod quic_config;

use std::{mem, sync::Arc, fmt::Write};

use tokio::{net, time::{self, Duration}, sync::mpsc};
use quinn::{Endpoint, Connection, ConnectionError, TransportErrorCode};

use quic_config::SecurityPolicy;

use super::message::{Message, WorldMessage, ScoreMessage};
use super::{Address, message_stream::{singleplayer::SingleplayerMessageStream, quic_client::QuicClientMs, DynMs, MsError}};

use crate::world::{World, MAX_STEPS, update::WorldUpdate};
use crate::scores::Scores;
use crate::blocks::Blocks;

pub struct Client {
	id: u32,
	updates: Vec<(WorldUpdate, u64)>,
	scores: Scores,
	stream: Box<DynMs>,
	sanity_checks: bool,
	sanity_check_world: Option<World>,
}

/**
 * An enum categorising types of errors methods of the Client return.
 */
pub enum ClientError {
	Io(MsError), // I/O errors in the message stream
	Protocol(String), // Protocol violations
	Limit(String), // When a limit to prevent application layer denial of service attacks is violated
}

pub type ClientRes = (u32, Client, World, Blocks);

impl Client {
	pub async fn multiplayer(endpoint: Result<Endpoint, String>, addr: Address, sanity_checks: bool) -> Result<ClientRes, String> {
		time::timeout(Duration::from_secs(10), async {
			let connection = Client::connect(endpoint?, addr.host.as_ref(), addr.port).await?;
			let (stream, init_msg) = QuicClientMs::build(connection, addr.game_id).await?;
			Client::build(init_msg, Box::new(stream), sanity_checks)
		}).await.unwrap_or_else(|_| Err(String::from("connection timed out")))
	}

	pub(super) async fn connect(endpoint: Endpoint, host: &str, port: u16) -> Result<Connection, String> {
		let mut policy = SecurityPolicy::new(host);
		let insecure_config = quic_config::new_insecure()?;
		let secure_config = quic_config::new_secure()?;

		let (conn_sender, mut conn_receiver) = mpsc::channel(16);

		// Performs DNS resolution if a domain name is provided
		let socket_addrs = net::lookup_host((host, port)).await.map_err(|err| format!("failed DNS resolution of host \"{host}\": {err}"))?.collect::<Vec<_>>();
		let host: Arc<str> = Arc::from(host);
		let mut conn_tasks = Vec::with_capacity(socket_addrs.len());

		/*
		 * Attempts to connect to every result from DNS resolution in parallel.
		 *
		 * To see the purpose of this, suppose that the server is bound to
		 * 127.0.0.1 and the user connects to `localhost`. This hostname first
		 * resolves to `::1` and then `127.0.0.1`. If this is done in parallel,
		 * then a lot of time will be spent waiting for a connection to `::1`,
		 * before nearly immediately connecting to `127.0.0.1` once the first
		 * connection times out.
		 */
		for socket_addr in socket_addrs {
			let secure = policy.needs_security(socket_addr.ip());
			let config = if secure { secure_config.clone() } else { insecure_config.clone() };

			let conn_sender = conn_sender.clone();

			/*
			 * When connecting to an IPv6 link-local address (fe80::/10), the scope
			 * id is important to resolve the ambiguity of which interface to use.
			 * However, scope ids are expressed as a '%' in the address, which
			 * isn't allowed for the server name, so I'm using any other address
			 * that works (in this case `localhost`).
			 */
			let server_name = if secure { &host } else { "localhost" };
			let connection_res = endpoint.connect_with(config, socket_addr, server_name);

			let handle = tokio::spawn(async move {
				match connection_res {
					Ok(connecting) => {
						let res = connecting.await.map_or_else(
							|err| {
								if let ConnectionError::TransportError(err) = &err {
									/*
									 * Provides a custom error message if the CA isn't
									 * recognised telling the user how to bypass the
									 * certificate check by manually entering the IP
									 * address.
									 *
									 * Note that TLS alert code 48 means the CA is
									 * unknown in both TLS 1.2 and 1.3: RFC 8446 (page
									 * 86) and RFC 5246 (page 69).
									 */
									if err.code == TransportErrorCode::crypto(48) {
										return Err((socket_addr.ip(), format!("unknown certificate authority: if you still want to connect to the server, manually enter the IP address of `{}` to bypass the certificate check. WARNING: Doing this makes you vulnerable to active eavesdropping and impersonation attacks (similar to browsing the web without using HTTPS).", socket_addr.ip())));
									}
								}

								Err((socket_addr.ip(), err.to_string()))
							},
							|connection| Ok((connection, secure)),
						);

						conn_sender.send(res).await
					}
					Err(err) => conn_sender.send(Err((socket_addr.ip(), err.to_string()))).await,
				}
			});

			conn_tasks.push(handle);
		}

		drop(conn_sender);

		// Iterates through all connections and uses the first one
		let mut errors = Vec::new();
		let connection = loop {
			match conn_receiver.recv().await {
				Some(Ok((conn, secure))) => {
					log::debug!("connecting to {host}, secure = {secure}");
					break conn;
				},
				Some(Err(err)) => errors.push(err),
				None if errors.is_empty() => return Err(format!("failed connecting to \"{host}\", no socket adresses were resolved for domain name")),
				None if errors.len() == 1 => return Err(format!("failed connecting to \"{host}\": {}", errors.last().map_or("[no error returned, something went wrong]", |err| err.1.as_str()))),
				None => {
					let mut err = format!("failed connecting to \"{host}\", {} socket addresses resolved and attempted for domain name, errors were: ", conn_tasks.len());
					for (i, e) in errors.into_iter().enumerate() {
						if i != 0 { err.push_str(", "); }
						let _ = write!(err, "\"{}\" from IP address `{}`", e.1, e.0);
					}
					return Err(err);
				}
			}
		};

		// Cancels all tasks to prevent any connections still attempted from wasting resources
		for task in conn_tasks {
			task.abort();
		}

		Ok(connection)
	}

	pub fn singleplayer(sanity_checks: bool) -> ClientRes {
		let (init_msg, stream) = SingleplayerMessageStream::new(sanity_checks);
		Client::build(init_msg, Box::new(stream), sanity_checks).unwrap()
	}

	fn build(init_msg: Message, stream: Box<DynMs>, sanity_checks: bool) -> Result<ClientRes, String> {
		// Receives the response to connecting
		let Message::World(WorldMessage::InitWorld { player_id: id, world, blocks: compressed_blocks }) = init_msg else {
			return Err(String::from("expected InitWorld message, got something else"));
		};

		if !world.has_player(id) { return Err(String::from("player not in world")); }

		let client = Client {
			id,
			updates: Vec::new(),
			scores: Scores::default(),
			stream,
			sanity_checks,
			sanity_check_world: None,
		};

		Ok((id, client, world, Blocks::try_from(compressed_blocks)?))
	}

	pub fn get_scores(&self) -> &Scores { &self.scores }

	pub fn is_networked(&self) -> bool {
		self.stream.is_networked()
	}

	pub fn send(&mut self, msg: Message) -> Result<(), ClientError> {
		self.stream.send(msg).map_err(ClientError::Io)
	}

	pub fn flush(&mut self) -> Result<(), ClientError> {
		self.stream.flush().map_err(ClientError::Io)
	}

	pub fn process_messages(&mut self) -> Result<(), ClientError> {
		loop {
			match self.stream.receive() {
				Ok(Some(Message::World(WorldMessage::WorldUpdate(update, ack)))) => {
					if !update.players.contains_key(&self.id) {
						return Err(ClientError::Protocol(String::from("sending an event without this player")));
					}

					self.updates.push((update, ack));

					if self.updates.len() > MAX_STEPS { // Client has been inactive for a while
						return Err(ClientError::Limit(String::from("too many steps accumulated")));
					}

					if self.sanity_check_world.take().is_some() {
						log::debug!("ignoring sanity check, received another world event"); // The world sent by the server is outdated, so this must be ignored
					}
				},
				Ok(Some(Message::Score(ScoreMessage::InitScores(scores)))) => {
					self.scores = scores;
					self.scores.validate()?;
				},
				Ok(Some(Message::Score(ScoreMessage::Score(id, score)))) => {
					self.scores.set(id, score);
					self.scores.validate()?;
				},
				Ok(Some(Message::World(WorldMessage::SanityCheck(world)))) => {
					if self.sanity_checks {
						self.sanity_check_world = Some(world);
					}
				},
				Ok(Some(_)) => return Err(ClientError::Protocol(String::from("sending an invalid message"))),
				Ok(None) => return Ok(()),
				Err(err) => return Err(ClientError::Io(err)),
			}
		}
	}

	pub fn get_world_updates(&mut self) -> Vec<(WorldUpdate, u64)> { mem::take(&mut self.updates) }
	pub fn get_sanity_check_world(&mut self) -> Option<World> { self.sanity_check_world.take() }
}
