// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use quinn::{Connection, SendStream, RecvStream};
use tokio::io::{BufReader, BufWriter};

use super::{MessageStream, DynMs, MsError, quic_common::{SenderTask, ReceiverTask, ReceiverTaskBuilder, ErrorContainer, MessageSer, MessageDe}};

use crate::net::{message::{Message, WorldMessage, ScoreMessage}, server::{SHUTTING_DOWN_MSG, SHUTTING_DOWN_STATUS}, serp::{Response, Vu30}, connection_listener::IncomingConnection};
use crate::player::update::{PlayerUpdate, PlayerBulkUpdate};
use crate::utils::net::SocketAddrToCanonical;

pub struct QuicServerMsIncoming {
	stream: QuicServerMs,
	subrequest_index: Vu30,
}

pub struct QuicServerMs {
	world_sender: SenderTask<WorldMessage>,
	world_receiver: ReceiverTask,
	score_sender: SenderTask<ScoreMessage>,
	connection: Connection,
	error: ErrorContainer,
}

impl QuicServerMsIncoming {
	pub async fn build(connection: Connection, sender_stream: BufWriter<SendStream>, receiver_stream: BufReader<RecvStream>, subrequest_index: Vu30) -> Result<QuicServerMsIncoming, String> {
		sender_stream.get_ref().set_priority(1).map_err(|err| format!("failed setting priority of world stream: {err}"))?;

		let error = ErrorContainer::default();
		let world_sender = SenderTask::<WorldMessage>::new(sender_stream, error.clone());
		let world_receiver_builder = ReceiverTaskBuilder::new(error.clone(), 1 << 8).add_stream::<(PlayerUpdate, u64)>(receiver_stream);
		let datagram_sender = world_receiver_builder.get_sender();
		let world_receiver = world_receiver_builder.build();

		let score_stream = connection.open_uni().await.map_err(|err| format!("failed creating unidirectional score stream: {err}"))?;
		let score_sender = SenderTask::<ScoreMessage>::new(BufWriter::new(score_stream), error.clone());

		{
			let error = error.clone();
			let connection = connection.clone();
			tokio::spawn(async move {
				while let Ok(data) = connection.read_datagram().await {
					match PlayerBulkUpdate::deserialise(&data) {
						Ok(bulk_update) => _ = datagram_sender.send(Message::PlayerBulkUpdate(bulk_update)).await,
						Err(err) => {
							error.insert_if_empty(MsError::Other(format!("failed deserialising bulk update: {err}")));
							return;
						},
					}
				}
			});
		}

		let stream = QuicServerMs { world_sender, world_receiver, score_sender, connection, error };
		Ok(QuicServerMsIncoming { stream, subrequest_index })
	}

	fn send_serp_response(&self, response: Response) -> Result<(), String> {
		self.stream.world_sender.send_serp_response(response, &self.stream.error, &self.stream.connection).map_err(|err| err.to_string())
	}

	fn flush_serp_response(&self) {
		let _ = self.stream.world_sender.flush(&self.stream.error, &self.stream.connection);
	}
}

impl IncomingConnection for QuicServerMsIncoming {
	fn accept(self: Box<Self>) -> Result<Box<DynMs>, (String, String)> {
		self.send_serp_response(Response::Success(self.subrequest_index)).map_err(|err| (err, self.stream.connection.remote_address().to_string()))?;
		// No point flushing it as the server's response is only useful for the client when they get the initial Messages
		Ok(Box::new(self.stream))
	}

	fn reject(self: Box<Self>, reason: String) {
		let _ = self.send_serp_response(Response::Understood(self.subrequest_index, reason));
		self.flush_serp_response();
	}
}

impl MessageStream for QuicServerMs {
	fn send(&mut self, msg: Message) -> Result<(), MsError> {
		match msg {
			Message::World(msg) => self.world_sender.send(msg, &self.error, &self.connection),
			Message::Score(msg) => self.score_sender.send(msg, &self.error, &self.connection),
			_ => unimplemented!(),
		}
	}

	fn receive(&mut self) -> Result<Option<Message>, MsError> {
		match self.error.get() { // Manually need to check when receiving as receiving only fails when both streams fail, so one can fail and that won't be reported
			Some(error) => Err(error),
			None => self.world_receiver.receive(&self.error, &self.connection),
		}
	}

	fn flush(&mut self) -> Result<(), MsError> {
		self.world_sender.flush(&self.error, &self.connection)?;
		self.score_sender.flush(&self.error, &self.connection)
	}

	#[cfg(feature = "client")]
	fn is_networked(&self) -> bool {
		true
	}

	fn close(self: Box<Self>) {
		self.connection.close(SHUTTING_DOWN_STATUS, SHUTTING_DOWN_MSG);
	}

	fn source(&self) -> String {
		/*
		 * This can change over time if the client ever changes networks (or NAT
		 * stuff) but it most likely will be the same as the reported address in
		 * the logs.
		 */
		self.connection.remote_address().to_canonical().to_string()
	}
}

impl MessageSer for WorldMessage {}
impl MessageSer for ScoreMessage {}

impl MessageDe for (PlayerUpdate, u64) {
	fn into_message(self) -> Message {
		Message::PlayerUpdate(self.0, self.1)
	}
}
