// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod fill_holes;
mod value_noise;
mod maze;
mod load_image;
mod vm;
#[cfg(test)]
mod tests;

use std::num::{NonZeroU32, NonZeroUsize};
use std::path::Path;

use serde::Deserialize;
use glam::{UVec2, IVec2, Vec2};
use strum::IntoEnumIterator;

use vm::Vm;

use crate::world::World;
use crate::player::ammo_count::AmmoCount;
use crate::blocks::Blocks;
use crate::powerup::Effect;

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) struct Config {
	#[serde(default = "max_players_default")] pub max_players: Option<NonZeroUsize>,
	pub world_size: Vec2,
	#[serde(default)] pub lan_discovery: LanDiscoveryConfig,
	pub blocks: BlockConfig,
	#[serde(default)] pub events: EventConfig,
}

#[derive(Default, Deserialize)]
#[serde(rename = "LanDiscovery", deny_unknown_fields)]
pub(in crate::net) struct LanDiscoveryConfig {
	#[serde(default)] pub name: String,
	#[serde(default)] pub desc: String,
}

#[derive(Deserialize)]
#[serde(rename = "Blocks", deny_unknown_fields)]
pub(in crate::net) struct BlockConfig {
	pub block_size: NonZeroU32,
	pub grid_size: UVec2,
	pub ops: Box<[Op]>,
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) enum Op {
	// Binary operations
	Add,
	Sub,
	Mul,
	Div,
	Min,
	Max,

	// Generation
	Constant(f32),
	WhiteNoise { #[serde(default)] seed: u64 },
	ValueNoise {
		#[serde(default)] seed: u64,
		levels: ValueNoiseLevels,
	},
	Maze {
		#[serde(default)] seed: u64,
		#[serde(default = "nonzerou32_one")] wall_thickness: NonZeroU32,
		#[serde(default = "nonzerou32_one")] open_thickness: NonZeroU32,
		#[serde(default)] open_prob: f32,
		#[serde(default)] close_prob: f32,
	},
	Circle(Vec2),
	Rect {
		pos: IVec2,
		size: IVec2,
		#[serde(default = "f32_one")] fill: f32,
	},
	Image(Box<Path>),

	// Utilities
	FillHoles,
	Threshold,

	// Stack operations
	Duplicate(isize),
	Remove(isize),
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) enum ValueNoiseLevels {
	Fractal {
		min: NonZeroU32,
		scale: NonZeroU32,
		octaves: NonZeroU32,
	},
	Sizes(Box<[NonZeroU32]>),
	WeightedSizes(Box<[(NonZeroU32, f32)]>),
}

#[derive(Deserialize)]
#[serde(rename = "Events", deny_unknown_fields)]
pub struct EventConfig {
	/**
	 * Radius of the square in which the players can spawn, respawn and teleport
	 * to.
	 */
	#[serde(default = "player_spawn_radius_default")] pub player_spawn_radius: f32,
	#[serde(default)] pub ammo_count: AmmoCountIo,
	#[serde(default = "ammo_crate_supply_range_default")] pub ammo_crate_supply_range: (u32, u32),

	/**
	 * There is a fixed probability of an item spawning in a given update,
	 * independent of the previous spawned items, but proportional to the number
	 * of players.
	 *
	 * These variables here are the mean time period for an item to spawn. To
	 * derive the probability of an item spawning in an update, divide delta time
	 * by this and multiply it by the number of players in the game.
	 */
	#[serde(default = "ammo_crate_period_default")] pub ammo_crate_period: f32,
	#[serde(default = "powerup_period_default")] pub powerup_period: f32,

	/**
	 * When an item is spawned, a player is selected at random and the item is
	 * spawned within a square of this radius centred at that player.
	 */
	#[serde(default = "item_spawn_radius_default")] pub item_spawn_radius: f32,

	/**
	 * Specifies a list of effects that powerups are allowed to be.
	 *
	 * When a powerup is spawned, a randomly selected effect is chosen from this
	 * list, which means that if the list contains duplicates, those duplicates
	 * are weighted higher.
	 *
	 * If this list is empty, no powerups are ever spawned.
	 *
	 * By default this is the list of all effects.
	 */
	#[serde(default = "allowed_effects_default")] pub allowed_effects: Box<[Effect]>,
}

impl Default for EventConfig {
	fn default() -> EventConfig {
		EventConfig {
			player_spawn_radius: player_spawn_radius_default(),
			ammo_count: AmmoCountIo::default(),
			ammo_crate_supply_range: ammo_crate_supply_range_default(),
			ammo_crate_period: ammo_crate_period_default(),
			powerup_period: powerup_period_default(),
			item_spawn_radius: item_spawn_radius_default(),
			allowed_effects: allowed_effects_default(),
		}
	}
}

#[derive(Deserialize)]
#[serde(from = "f64", deny_unknown_fields /* Don't think it's needed */)]
pub struct AmmoCountIo(pub AmmoCount);

impl From<f64> for AmmoCountIo {
	fn from(count: f64) -> AmmoCountIo {
		AmmoCountIo(if count == f64::INFINITY {
			AmmoCount::new_infinite()
		} else {
			AmmoCount::new(count as u32)
		})
	}
}

impl Default for AmmoCountIo {
	fn default() -> AmmoCountIo {
		AmmoCountIo(AmmoCount::new(100))
	}
}

// It would be nice if serde could inline this in the attribute
fn nonzerou32_one() -> NonZeroU32 { NonZeroU32::new(1).unwrap() }
fn f32_one() -> f32 { 1.0 }
fn ammo_crate_period_default() -> f32 { 30.0 }
fn item_spawn_radius_default() -> f32 { 40.0 }
fn powerup_period_default() -> f32 { 60.0 }
fn player_spawn_radius_default() -> f32 { 64.0 }
fn ammo_crate_supply_range_default() -> (u32, u32) { (50, 75) }
fn allowed_effects_default() -> Box<[Effect]> { Effect::iter().collect::<Box<_>>() }
fn max_players_default() -> Option<NonZeroUsize> { NonZeroUsize::new(8) }

pub(super) type BlockMap = Box<[f32]>;

impl Config {
	pub fn generate(self) -> Result<(World, Blocks, EventConfig), String> {
		let world = World::new(self.world_size);
		let block_map = self.generate_block_map()?;
		let blocks = Blocks::new(self.blocks.block_size, self.blocks.grid_size, block_map.iter().map(|&x| x > 0.0).collect());

		if self.events.ammo_crate_supply_range.0 >= self.events.ammo_crate_supply_range.1 {
			return Err(format!("ammo crate supply range (a, b) requires a ≤ b, got (a, b) = {}, {}", self.events.ammo_crate_supply_range.0, self.events.ammo_crate_supply_range.1));
		}

		Ok((world, blocks, self.events))
	}

	pub fn generate_block_map(&self) -> Result<BlockMap, String> {
		Vm::new(self.blocks.grid_size).generate(&self.blocks.ops)
	}
}
