// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::ops::{Add, Sub, AddAssign};

use glam::Vec2;

use super::{Player, direction::Angle};

#[derive(Default)]
pub struct DPlayer {
	pos: Vec2,
	vel: Vec2,
	angle: Angle,
}

impl DPlayer {
	pub fn zero() -> DPlayer { DPlayer::default() }

	pub fn get_pos(&self) -> Vec2 { self.pos }
	pub fn get_vel(&self) -> Vec2 { self.vel }
	pub fn get_angle(&self) -> Angle { self.angle }

	pub fn decay(&mut self, amount: f32) {
		/*
		 * Decays the direction faster as a jump in the position can feel very
		 * weird, but a jump in the direction can be normal gameplay.
		 */
		let (scale, scale_dir) = ((1.0 - amount).max(0.0), (1.0 - amount * 8.0).max(0.0));

		self.pos *= scale;
		self.vel *= scale;
		self.angle *= scale_dir;
	}
}

impl Sub for &Player {
	type Output = DPlayer;

	fn sub(self, other: &Player) -> DPlayer {
		DPlayer {
			pos: self.pos - other.pos,
			vel: self.vel - other.vel,
			angle: self.dir - other.dir,
		}
	}
}

impl Add<&DPlayer> for &Player {
	type Output = Player;

	fn add(self, delta: &DPlayer) -> Player {
		Player {
			pos: self.pos + delta.pos,
			vel: self.vel + delta.vel,
			dir: self.dir + delta.angle,
			.. self.clone()
		}
	}
}

impl AddAssign<&DPlayer> for DPlayer {
	fn add_assign(&mut self, rhs: &DPlayer) {
		self.pos += rhs.pos;
		self.vel += rhs.vel;
		self.angle += rhs.angle;
	}
}
