// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::fmt::Write;

use crate::net::{message::Message, client::{Client, ClientRes, ClientError}};
use crate::world::{World, changes::ClientChanges};
use crate::blocks::Blocks;

use super::TimeArgs;
use super::view::View;
use super::controller::Controller;

pub struct Model {
	pub(super) id: u32,
	pub(super) client: Client,
	pub(super) world: World,
	pub(super) blocks: Blocks,
}

impl Model {
	pub(super) fn new(client_res: ClientRes) -> Model {
		let (id, client, world, blocks) = client_res;
		Model { id, client, world, blocks }
	}

	pub fn get_id(&self) -> u32 { self.id }
	pub fn get_world(&self) -> &World { &self.world }
	pub fn get_blocks(&self) -> &Blocks { &self.blocks }

	pub(super) fn update(&mut self, view: &mut View, controller: &mut Controller, time: &TimeArgs) -> Result<(), ClientError> {
		self.client.process_messages()?;

		let init_player_pos = view.smooth_world.get_player_pos(time.time_diff);
		controller.pcontrol.mouse_move_event(view.camera.pixel_to_world_coords(controller.cursor_pos) - init_player_pos);
		let player_update = controller.pcontrol.get_update();

		let updates = self.client.get_world_updates();
		if !updates.is_empty() {
			let mut changes = ClientChanges::default();
			for update in &updates {
				self.world.apply_events(&update.0.pre_events, &mut changes);
				self.world.update(&update.0.players, &self.blocks, &mut changes);
				self.world.apply_events(&update.0.post_events, &mut changes);
				self.world.validate().map_err(|err| ClientError::Limit(String::from(err)))?;
			}

			let ack = updates.last().unwrap().1;
			view.smooth_world.sync(&self.world, &self.blocks, &changes, ack, updates.len());

			for hit in &changes.hits {
				view.on_hit(hit);
			}

			if changes.moved.contains(&self.id) {
				view.camera.move_by(view.smooth_world.get_player_pos(time.time_diff) - init_player_pos);
			}
		}

		if time.steps > 0 {
			for _ in 0..time.steps {
				view.smooth_world.update(player_update, &self.world, &self.blocks)?;
			}

			self.client.send(Message::PlayerUpdate(player_update, time.steps as u64))?;
			if self.client.is_networked() {
				self.client.send(view.smooth_world.create_bulk_update())?;
			}
			self.client.flush()?;
		}

		Ok(())
	}

	pub(super) fn sanity_check(&mut self) {
		if let Some(sanity_check_world) = self.client.get_sanity_check_world() {
			let mut expected_str = String::new();
			write!(&mut expected_str, "{sanity_check_world:?}").unwrap();
			let mut actual_str = String::with_capacity(expected_str.len());
			write!(&mut actual_str, "{:?}", &self.world).unwrap();

			if expected_str != actual_str {
				log::error!("world sync error (id = {}):\nexpected = {expected_str},\nactual = {actual_str}", self.id);
			} else {
				log::debug!("sanity check passed");
			}
		}
	}
}
