// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{rc::Rc, cell::RefCell};

use glium::{Display, Texture2d, Frame, DrawParameters};
use glutin::surface::WindowSurface;
use glam::{Vec2, Mat4};

use super::{ALPHA_BLENDING, ADDITIVE_BLENDING};

use crate::game::{config::Config, filesystem::Filesystem};
use crate::camera::Camera;
use crate::player::death_effect::DeathEffect;
use crate::stars::Stars;
use crate::world::{SmoothWorld, changes::Hit as ChangesHit};
use crate::particle::{ParticleSystem, hit::{Hit, HitArgs}, thrust::Thrust};
use crate::powerup::{Powerup, particle_clock::ParticleClock};
use crate::utils::{Clock, maths::Circle};

pub(super) struct Decorations {
	config: Rc<RefCell<Config>>,

	stars: Stars,
	particles: ParticleSystem,
	death_effect: DeathEffect,

	powerup_particle_clock: ParticleClock,
	thrust_particle_clock: Clock,
}

impl Decorations {
	pub fn new(config: Rc<RefCell<Config>>, display: &Display<WindowSurface>, fs: &Filesystem, player_texture: Rc<Texture2d>) -> Decorations {
		Decorations {
			config,
			stars: Stars::new(display, fs),
			particles: ParticleSystem::new(display, fs),
			death_effect: DeathEffect::new(display, fs, player_texture),
			powerup_particle_clock: ParticleClock::new(),
			thrust_particle_clock: Clock::new(1.0 / 200.0),
		}
	}

	pub fn reset(&mut self, camera: &Camera) {
		self.stars.reset(camera, self.config.borrow().graphics.star_density);
		self.particles.reset();
		self.death_effect.reset();
	}

	pub fn on_hit(&mut self, hit: &ChangesHit) {
		let particle_config = &self.config.borrow().graphics.particles;
		let args = HitArgs::new(hit);

		if particle_config.hit {
			self.particles.add(|| Hit::new(hit, &args), 8);
		}

		if particle_config.death && hit.killed {
			self.death_effect.add(hit, &mut self.particles);
		}
	}

	pub fn render_stars(&mut self, dt: f32, display: &Display<WindowSurface>, camera: &Camera, to_move: Vec2, frame: &mut Frame, params: &mut DrawParameters) {
		self.stars.update(dt, camera, to_move);
		params.multisampling = false; // Disables multisampling as that results in some points flickering as they move across the screen, which can be distracting
		self.stars.render(display, frame, params);
		params.multisampling = true;
	}

	pub fn update_particles(&mut self, dt: f32) {
		self.particles.update(dt);
	}

	pub fn add_particles(&mut self, dt: f32, time_diff: f32, smooth_world: &SmoothWorld, powerups: &[Powerup]) {
		let (thrust, powerup) = {
			let particle_config = &self.config.borrow().graphics.particles;
			(particle_config.thrust, particle_config.powerup)
		};

		if thrust { self.add_thrust_particles(dt, time_diff, smooth_world); }
		if powerup { self.add_powerup_particles(dt, powerups); }
	}

	fn add_thrust_particles(&mut self, dt: f32, time_diff: f32, smooth_world: &SmoothWorld) {
		let to_add = self.thrust_particle_clock.advance(dt);
		if to_add == 0 { return; }

		for (_, player) in smooth_world.player_iter() {
			if player.inner().get_input_state().thrusting() {
				let pos = player.get_pos(time_diff);
				let vel = player.get_vel();
				let dir = player.get_dir();
				self.particles.add(|| Thrust::new(pos, vel, dir), to_add as usize);
			}
		}

	}

	fn add_powerup_particles(&mut self, dt: f32, powerups: &[Powerup]) {
		if self.powerup_particle_clock.advance(dt) {
			for powerup in powerups {
				let effect = powerup.get_effect();
				self.particles.add(|| effect.new_particle(powerup.get_pos()), self.powerup_particle_clock.get_count(effect) as usize);
			}
		}
	}

	pub fn render_particles(&mut self, dt: f32, display: &Display<WindowSurface>, frame: &mut Frame, params: &mut DrawParameters, matrix: &Mat4) {
		// Uses additive blending for the particles to avoid a white particle being rendered over by a darker particle, resulting in things appearing less bright
		params.blend.color = ADDITIVE_BLENDING;
		self.particles.render(display, frame, params, matrix);
		params.blend.color = ALPHA_BLENDING;

		self.death_effect.update(dt);
		self.death_effect.render(display, frame, params, matrix);
	}
}
