// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use glium::{Display, Frame, DrawParameters};
use glutin::surface::WindowSurface;
use glam::{UVec2, Vec2};

use crate::game::{Game, filesystem::Filesystem};
use crate::ammo_status::AmmoStatus;
use crate::health_bar::HealthBarRenderer;
use crate::world::{World, SmoothWorld};
use crate::camera::Camera;
use crate::scores::{self, Scores};

pub(super) struct Hud {
	ammo_status: AmmoStatus,
	health_bars: HealthBarRenderer,
	winsize: UVec2,
	scale_factor: f32,
}

impl Hud {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem, winsize: UVec2, scale_factor: f32) -> Hud {
		Hud {
			ammo_status: AmmoStatus::new(display, fs),
			health_bars: HealthBarRenderer::new(display, fs),
			winsize,
			scale_factor,
		}
	}

	pub fn reset(&mut self, winsize: UVec2, scale_factor: f32) {
		self.resize_event(winsize);
		self.scale_factor_event(scale_factor);
	}

	pub fn resize_event(&mut self, winsize: UVec2) { self.winsize = winsize; }
	pub fn scale_factor_event(&mut self, scale_factor: f32) { self.scale_factor = scale_factor; }

	pub fn push_health_bars(&mut self, smooth_world: &SmoothWorld, camera: &Camera, time_diff: f32) {
		let lwinsize = self.get_lwinsize();
		smooth_world.push_health_bars(&mut self.health_bars, lwinsize, camera, time_diff);
	}

	pub fn render(&mut self, game: &mut Game, frame: &mut Frame, params: &mut DrawParameters, world: &World, id: u32, scores: &Scores) {
		let lwinsize = self.get_lwinsize();

		self.health_bars.render(&game.display, frame, params);
		self.ammo_status.render(frame, params, lwinsize);
		self.ammo_status.set_text(world.get_ammo_string(id));

		params.multisampling = false;
		game.gui.update(&game.window, |ctx| {
			/*
			 * Temporarily sets the animation time to zero so that the ammo status
			 * window appears immediately.
			 */
			let mut prev_animation_time = 0.0;
			ctx.style_mut(|style| {
				prev_animation_time = style.animation_time;
				style.animation_time = 0.0;
			});

			self.ammo_status.render_text(ctx, lwinsize);
			scores::renderer::render(ctx, scores, id);

			ctx.style_mut(|style| style.animation_time = prev_animation_time);
		});

		game.gui.render(&game.display, frame);
		params.multisampling = true;
	}

	fn get_lwinsize(&self) -> Vec2 {
		self.winsize.as_vec2() / self.scale_factor
	}
}
