// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
#[cfg(feature = "client")] mod renderer;
#[cfg(feature = "client")] pub mod particle_clock;
pub mod active_effects;

#[cfg(feature = "client")] pub use renderer::Renderer as PowerupRenderer;

use strum_macros::{EnumCount as EnumCountMacro, EnumIter};
use serde::{Serialize, Deserialize};
use glam::Vec2;

use crate::utils::maths::Circle;
#[cfg(feature = "client")] use crate::particle::{AnyParticle, powerup::{speed::Speed, reload::Reload, health::Health, damage::Damage, forcefield::Forcefield, teleportation::Teleportation}};

#[derive(Debug, Clone, Copy, PartialOrd, PartialEq, Eq, Serialize, Deserialize, EnumCountMacro, EnumIter)]
pub enum Effect {
	// Temporal effects
	Speed,
	Reload,
	Health,
	Damage,
	Forcefield,

	// Instantaneous effects
	Teleportation,
}

const FIRST_INSTANTANEOUS: Effect = Effect::Teleportation;

#[derive(PartialEq, Eq)]
pub enum EffectType {
	/** An effect that lasts for some period of time. */
	Temporal,

	/** An effect that happens instantly. */
	Instantaneous,
}

impl Effect {
	pub fn get_type(self) -> EffectType {
		if self < FIRST_INSTANTANEOUS { EffectType::Temporal }
		else { EffectType::Instantaneous }
	}

	#[cfg(feature = "client")]
	pub fn new_particle(self, pos: Vec2) -> AnyParticle {
		match self {
			Effect::Speed => Speed::new(pos).into(),
			Effect::Reload => Reload::new(pos).into(),
			Effect::Health => Health::new(pos).into(),
			Effect::Damage => Damage::new(pos).into(),
			Effect::Forcefield => Forcefield::new(pos).into(),
			Effect::Teleportation => Teleportation::new(pos).into(),
		}
	}
}

pub const RADIUS: f32 = 1.0 / 3.0;
const TEMPORAL_EFFECT_COUNT: u32 = FIRST_INSTANTANEOUS as u32;
const MAX_TIME: f32 = 10.0 * 60.0;
const FADE_IN_TIME: f32 = 1.0;
const FADE_OUT_TIME: f32 = 30.0;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Powerup {
	pos: Vec2,
	scale: f32,
	time: f32,
	effect: Effect,
}

impl Powerup {
	pub fn new(pos: Vec2, effect: Effect) -> Powerup {
		Powerup { pos, scale: 0.0, time: 0.0, effect }
	}

	pub fn get_effect(&self) -> Effect { self.effect }

	pub fn update(&mut self, dt: f32) -> bool {
		self.time += dt;
		self.scale = Powerup::get_scale(self.time);
		self.time <= MAX_TIME
	}

	fn get_scale(time: f32) -> f32 {
		if time < FADE_IN_TIME {
			let f = time / FADE_IN_TIME;
			f * (2.0 - f)
		} else if time > MAX_TIME - FADE_OUT_TIME {
			let f = (MAX_TIME - time) / FADE_OUT_TIME;
			f * (2.0 - f)
		} else {
			1.0
		}
	}
}

impl Circle for Powerup {
	fn get_pos(&self) -> Vec2 { self.pos }
	fn get_radius(&self) -> f32 { RADIUS * self.scale }
}
