// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use serde::{Serialize, Deserialize};

use super::{Effect, TEMPORAL_EFFECT_COUNT};

#[cfg(feature = "client")] use crate::world::MAX_PLAYERS;
use crate::world::DMap;

#[derive(Debug, Default, Clone, Serialize, Deserialize)]
struct EffectList([f32; TEMPORAL_EFFECT_COUNT as usize]);

impl EffectList {
	fn has(&self, effect: Effect) -> bool { self.0[effect as usize] > 0.0 }
	fn get_time(&self, effect: Effect) -> f32 { self.0[effect as usize] }

	fn add(&mut self, effect: Effect) {
		let time = self.0.get_mut(effect as usize).unwrap();
		*time = time.max(0.0) + EFFECT_TIME;
	}

	fn update(&mut self, dt: f32) -> bool {
		let mut has_effects = false;
		for time in self.0.iter_mut().filter(|&&mut time| time > 0.0) {
			has_effects = true;
			*time -= dt;
		}
		has_effects
	}
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ActiveEffects(DMap<u32, EffectList>);

const EFFECT_TIME: f32 = 30.0;

impl ActiveEffects {
	pub fn new() -> ActiveEffects {
		ActiveEffects(DMap::new())
	}

	#[cfg(feature = "client")]
	pub fn validate(&self) -> Result<(), &'static str> {
		if self.0.len() <= MAX_PLAYERS { Ok(()) }
		else { Err("too many players") }
	}

	pub fn remove_player(&mut self, id: u32) {
		self.0.remove(&id);
	}

	pub fn get_mul(&self, id: u32, effect: Effect) -> f32 {
		if self.has_effect(id, effect) {
			match effect {
				Effect::Health => 2.0,
				Effect::Speed => 1.375,
				_ => 1.5,
			}
		} else { 1.0 }
	}

	pub fn add_effect(&mut self, id: u32, effect: Effect) {
		self.0.entry(id).or_default().add(effect);
	}

	/**
	 * Returns an iterator of all player ids and the corresponding time of all
	 * players with `effect`.
	 */
	pub fn player_iter_with_effect(&self, effect: Effect) -> impl Iterator<Item = (u32, f32)> + '_ {
		self.0.iter().filter(move |(_, effects)| effects.has(effect)).map(move |(&id, effects)| (id, effects.get_time(effect)))
	}

	fn has_effect(&self, id: u32, effect: Effect) -> bool {
		self.0.get(&id).map_or(false, |effects| effects.has(effect))
	}

	pub fn update(&mut self, dt: f32) {
		self.0.retain(|_, effects| effects.update(dt));
	}
}
