// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod glam_fix;

use glam::Vec2;

pub fn rotate_by(vec: Vec2, angle: f32) -> Vec2 {
	rotate_by_dir(vec, angle_to_dir(angle))
}

pub fn rotate_by_dir(vec: Vec2, dir: Vec2) -> Vec2 {
	Vec2::new(vec.x * dir.x - vec.y * dir.y, vec.x * dir.y + vec.y * dir.x) // Rotates the vector by some amount (uses the 2D rotation matrix)
}

pub fn angle_to_dir(angle: f32) -> Vec2 {
	let (sin, cos) = libm::sincosf(angle);
	Vec2::new(cos, sin)
}

/**
 * Linearly interpolates between `p0` and `p1` by `t`. Note that lerp(p0, p1,
 * 1) isn't guaranteed to be p1 due to possible floating-point error.
 */
pub fn lerp(p0: f32, p1: f32, t: f32) -> f32 {
	p0 + (p1 - p0) * t
}

pub fn bilerp(p00: f32, p10: f32, p01: f32, p11: f32, x: f32, y: f32) -> f32 {
	lerp(lerp(p00, p10, x), lerp(p01, p11, x), y)
}

pub trait Rect {
	fn get_p0(&self) -> Vec2;
	fn get_p1(&self) -> Vec2;
}

pub trait Circle {
	fn get_pos(&self) -> Vec2;
	fn get_radius(&self) -> f32;

	fn colliding_circle(&self, other: &impl Circle) -> bool {
		let sum_r = self.get_radius() + other.get_radius();
		self.get_pos().distance_squared(other.get_pos()) < sum_r * sum_r
	}

	/*
	 * Rectangle-circle collision detection algorithm. This works by checking if
	 * the circle's centre is inside the rectangle, or is outside it but within
	 * distance radius of it.
	 */
	fn colliding_rect(&self, other: &impl Rect) -> bool {
		// Using a wrapper to avoid unnecessary code size with monomorphisation
		fn colliding(pos: Vec2, radius: f32, p0: Vec2, p1: Vec2) -> bool {
			// Checks the first two rectangles
			let (suf_right, suf_left) = (pos.x > p0.x, pos.x < p1.x); // If sufficiently right or left
			if suf_right && suf_left && pos.y > p0.y - radius && pos.y < p1.y + radius {
				return true;
			}

			let (suf_up, suf_down) = (pos.y > p0.y, pos.y < p1.y);
			if suf_up && suf_down && pos.x > p0.x - radius && pos.x < p1.x + radius {
				return true;
			}

			// Now checking a corner
			// Using the previous results to avoid searching all four
			let corner = match (suf_right, suf_up) {
				(false, false) => p0,
				( true, false) => Vec2::new(p1.x, p0.y),
				(false,  true) => Vec2::new(p0.x, p1.y),
				( true,  true) => p1,
			};

			pos.distance_squared(corner) < radius * radius
		}

		colliding(self.get_pos(), self.get_radius(), other.get_p0(), other.get_p1())
	}
}

impl Circle for (Vec2, f32) {
	fn get_pos(&self) -> Vec2 { self.0 }
	fn get_radius(&self) -> f32 { self.1 }
}
