// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod udp;

use std::net::{SocketAddr, SocketAddrV4, IpAddr, Ipv4Addr, Ipv6Addr};

use sysinfo::{Networks, NetworkData};
use quinn::Endpoint;
use tokio::time::{self, Duration};

#[derive(Clone, Copy)]
pub struct Ipv4Network {
	pub addr: Ipv4Addr,
	pub net_prefix: u8,
}

impl Ipv4Network {
	pub fn networks(networks: &Networks) -> impl Iterator<Item = Ipv4Network> + '_ {
		networks
			.values()
			.flat_map(NetworkData::ip_networks)
			.filter_map(|ip_net| match ip_net.addr {
				IpAddr::V4(addr) => Some(Ipv4Network { addr, net_prefix: ip_net.prefix }),
				IpAddr::V6(_) => None,
			})
	}

	pub fn directed_broadcast(self) -> Ipv4Addr {
		Ipv4Addr::from_bits(self.addr.to_bits() | self.not_netmask())
	}

	pub fn contains(self, addr: Ipv4Addr) -> bool {
		let netmask = !self.not_netmask();
		(self.addr.to_bits() & netmask) == (addr.to_bits() & netmask)
	}

	fn not_netmask(self) -> u32 {
		let host_bits = u32::BITS - self.net_prefix as u32;
		(1 << host_bits) - 1
	}
}

#[derive(Clone, Copy)]
pub struct Ipv6Network {
	pub addr: Ipv6Addr,
	pub net_prefix: u8,
}

impl Ipv6Network {
	pub fn networks(networks: &Networks) -> impl Iterator<Item = Ipv6Network> + '_ {
		networks
			.values()
			.flat_map(NetworkData::ip_networks)
			.filter_map(|ip_net| match ip_net.addr {
				IpAddr::V4(_) => None,
				IpAddr::V6(addr) => Some(Ipv6Network { addr, net_prefix: ip_net.prefix }),
			})
	}

	pub fn contains(self, addr: Ipv6Addr) -> bool {
		let netmask = !self.not_netmask();
		(self.addr.to_bits() & netmask) == (addr.to_bits() & netmask)
	}

	fn not_netmask(self) -> u128 {
		let host_bits = u128::BITS as u128 - self.net_prefix as u128;
		(1 << host_bits) - 1
	}
}

pub trait SocketAddrToCanonical {
	fn to_canonical(self) -> SocketAddr;
}

impl SocketAddrToCanonical for SocketAddr {
	fn to_canonical(self) -> SocketAddr {
		if let IpAddr::V4(ip_addr) = self.ip().to_canonical() {
			SocketAddr::V4(SocketAddrV4::new(ip_addr, self.port()))
		} else {
			self
		}
	}
}

pub trait WaitIdleTimeout {
	async fn wait_idle_timeout(&self);
}

impl WaitIdleTimeout for Endpoint {
	async fn wait_idle_timeout(&self) {
		log::info!("Gracefully finishing any connections...");
		if time::timeout(Duration::from_secs(15), self.wait_idle()).await.is_err() {
			log::info!("Finishing connections timed out, exiting now");
		}
	}
}

pub trait NotPublic {
	fn ip_addr_not_public(&mut self, addr: IpAddr) -> bool {
		match addr.to_canonical() {
			IpAddr::V4(addr) => {
				if addr.is_loopback() || addr.is_private() || addr.is_link_local() { return true; }
				return Ipv4Network::networks(self.get_networks()).any(|ip_net| ip_net.contains(addr));
			},
			IpAddr::V6(addr) => {
				if addr.is_loopback() { return true; }

				for ip_net in [
					Ipv6Network { addr: Ipv6Addr::new(0xfe80, 0, 0, 0, 0, 0, 0, 0), net_prefix: 10 }, // Link-local
					Ipv6Network { addr: Ipv6Addr::new(0xfc00, 0, 0, 0, 0, 0, 0, 0), net_prefix: 7 }, // Private networks
				] {
					if ip_net.contains(addr) { return true; }
				}

				return Ipv6Network::networks(self.get_networks()).any(|ip_net| ip_net.contains(addr));
			},
		}
	}

	fn get_networks(&mut self) -> &Networks;
}
