// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod move_chooser;

use std::collections::BTreeSet;

use serde::{Serialize, Deserialize};
use glam::Vec2;
use rand::{SeedableRng, RngCore};

use move_chooser::MoveChooser;

use super::{World, WorldRng, changes::{Changes, ServerChanges}, DELTA_TIME, MAX_AMMO_CRATES, MAX_POWERUPS};

use crate::player::{Player, ammo_count::AmmoCount};
use crate::blocks::Blocks;
use crate::ammo_crate::{AmmoCrate, RADIUS as AMMO_CRATE_RADIUS};
use crate::net::server::config::EventConfig;
use crate::powerup::{Powerup, Effect, RADIUS as POWERUP_RADIUS};
use crate::utils::{rng::Float, maths::{self, Circle}};

#[derive(Clone, Serialize, Deserialize)]
pub enum WorldEvent {
	/// When a new player joins.
	NewPlayer, // No id included as that can be inferred from the current players (smallest unused)

	/// Sets the player's position.
	PlayerPos(u32, Vec2),

	/// Sets the player's velocity.
	PlayerVel(u32, Vec2),

	/// Sets the player's amount of ammo.
	PlayerAmmo(u32, AmmoCount),

	/// Sets the player's health.
	PlayerHealth(u32, f32),

	/// Adds a new ammo crate. The amount of ammo is determined when the player
	/// picks it up, in which the server may give the player ammo with the
	/// `PlayerAmmo` update.
	NewAmmoCrate(Vec2),

	/// Adds a new powerup.
	NewPowerup(Vec2, Effect),
}

pub struct EventChooser {
	rng: WorldRng,
	move_chooser: MoveChooser,
	config: EventConfig,
}

impl EventChooser {
	pub fn new(world: &World, blocks: &Blocks, config: EventConfig) -> EventChooser {
		let mut rng = WorldRng::seed_from_u64(rand::random::<u64>());
		let move_chooser = MoveChooser::new(world, blocks, config.player_spawn_radius, &mut rng);
		EventChooser { rng, move_chooser, config }
	}

	pub fn get_pre_events(&mut self, world: &World, blocks: &Blocks, new_players: Option<&BTreeSet<u32>>) -> Vec<WorldEvent> {
		/*
		 * Also using the client limits to prevent clients from disconnecting if
		 * there are too many. Not also doing this with bullets as that case is
		 * less likely to occur (this can occur by players staying idle for a
		 * while) and players not being able to shoot would be weird.
		 */
		let mut events = Vec::new();

		if self.rng.next_f32() < EventChooser::get_prob(self.config.ammo_crate_period, world.players.len()) && world.ammo_crates.len() < MAX_AMMO_CRATES {
			/*
			 * Note that my definition of collision is strictly overlapping (not
			 * just the edges touching), but iter_range has non-strict overlapping
			 * (even if the edges are touching it's counted as a collision). This
			 * means that some few edge (pun somewhat intended) cases can never
			 * have an ammo crate spawn when it's totally valid. This isn't a
			 * problem as those cases are very unlikely and any difference would be
			 * basically unnoticeable.
			 */
			if let Some(pos) = self.find_item_pos(world, AMMO_CRATE_RADIUS, |pos| blocks.iter_range(pos, AMMO_CRATE_RADIUS).count() == 0) {
				events.push(WorldEvent::NewAmmoCrate(pos));
			}
		}

		if self.rng.next_f32() < EventChooser::get_prob(self.config.powerup_period, world.players.len()) && world.powerups.len() < MAX_POWERUPS && !self.config.allowed_effects.is_empty() {
			if let Some(pos) = self.find_item_pos(world, POWERUP_RADIUS, |pos| !blocks.iter_range(pos, POWERUP_RADIUS).any(|block| (pos, POWERUP_RADIUS).colliding_rect(&block))) {
				let effect = self.config.allowed_effects[self.rng.next_u32() as usize % self.config.allowed_effects.len()];
				events.push(WorldEvent::NewPowerup(pos, effect));
			}
		}

		if let Some(new_players) = new_players {
			for &id in new_players {
				events.push(WorldEvent::NewPlayer);
				events.push(WorldEvent::PlayerPos(id, self.move_chooser.generate(blocks, &mut self.rng)));
				events.push(WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));
			}
		}

		events
	}

	pub fn get_post_events(&mut self, world: &World, blocks: &Blocks, changes: &ServerChanges) -> Vec<WorldEvent> {
		let mut events = Vec::new();

		for &id in &changes.randomly_moved {
			events.push(WorldEvent::PlayerPos(id, self.move_chooser.generate(blocks, &mut self.rng)));
		}

		for &id in &changes.dead {
			events.push(WorldEvent::PlayerVel(id, Vec2::ZERO));
			events.push(WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));
		}

		for &id in &changes.ammo_crates_collected {
			let range = self.config.ammo_crate_supply_range;
			let (min, max) = (range.0, range.1);
			let new_ammo = world.players[&id].get_ammo_count() + (min + self.rng.next_u32() % ((max - min).saturating_add(1)));
			events.push(WorldEvent::PlayerAmmo(id, new_ammo));
		}

		events
	}

	fn get_prob(mean_interval: f32, players: usize) -> f32 {
		(DELTA_TIME / mean_interval) * players as f32
	}

	fn find_item_pos(&mut self, world: &World, radius: f32, should_spawn: impl Fn(Vec2) -> bool) -> Option<Vec2> {
		debug_assert!(!world.players.is_empty()); // Nothing should spawn when there are no players

		for _ in 0..20 { // Gives up if no suitable place is found after too many tries
			// Lazy O(n) algorithm, might improve later
			let rand_player_pos = world.players.iter().nth(self.rng.next_u32() as usize % world.players.len()).unwrap().1.get_pos();

			// Gets the range for spawning, which must be inside the world
			let min = (rand_player_pos - Vec2::splat(self.config.item_spawn_radius)).max(radius - world.size / 2.0);
			let max = (rand_player_pos + Vec2::splat(self.config.item_spawn_radius)).min(world.size / 2.0 - radius);

			let pos = Vec2::new(maths::lerp(min.x, max.x, self.rng.next_f32()), maths::lerp(min.y, max.y, self.rng.next_f32()));

			// If not colliding with any block (or some other condition)
			if should_spawn(pos) {
				return Some(pos);
			}
		}

		None
	}
}

impl World {
	pub fn apply_events(&mut self, events: &[WorldEvent], changes: &mut impl Changes) {
		for event in events {
			match *event {
				WorldEvent::NewPlayer => self.new_player(),
				WorldEvent::PlayerPos(id, pos) => {
					self.player_present(id, |player| player.set_pos(pos));
					changes.pos_changed(id);
				},
				WorldEvent::PlayerVel(id, vel) => self.player_present(id, |player| player.set_vel(vel)),
				WorldEvent::PlayerHealth(id, health) => self.player_present(id, |player| player.set_health(health)),
				WorldEvent::PlayerAmmo(id, ammo) => self.player_present(id, |player| player.set_ammo_count(ammo)),
				WorldEvent::NewAmmoCrate(pos) => self.ammo_crates.push(AmmoCrate::new(pos)),
				WorldEvent::NewPowerup(pos, effect) => self.powerups.push(Powerup::new(pos, effect)),
			}
		}
	}

	fn player_present(&mut self, id: u32, f: impl Fn(&mut Player)) {
		if let Some(player) = self.players.get_mut(&id) {
			f(player);
		}
	}
}
